import numpy as np
import pandas as pd
import scipy.stats as stats
import statsmodels.api as sm
import sspa.utils as utils

"""
Module docstring
"""

class sspa_ora:
    """
    Class for overrepresentation analysis 

    Attributes:
        mat (pd.DataFrame): dataframe containing input metabolomics data
        metadata (pd.Series): series containing phenotype metadata e.g 'COVID', 'NON-COVID'
        pathways (pd.DataFrame): pathway dataframe containing compound identifiers
        DA_cutoff (float): pFDR cutoff for selecting differential metabolites e.g. 0.05 or 0.01
        DA_testtype (str): Test type for selecing differential metabolites. Can either be 'ttest' (default) for the independent students T-test or 'mwu' for the Mann Whitney U test, both as implemented in SciPy. 
        custom_background (list): background list of identifiers, default is to use annotated compounds in input data (i.e. mat.columns)
    """
    def __init__(self, mat, metadata, pathways, DA_cutoff, DA_testtype='ttest', custom_background=None):
        self.data = mat
        self.metadata = metadata
        self.pathways = pathways
        self.threshold = DA_cutoff
        self.testtype = DA_testtype
        self.background_set = custom_background if custom_background is not None else mat.columns.to_list()

        self.DA_test_res = utils.t_tests(self.data.copy(deep=True), self.metadata, "fdr_bh", testtype=self.testtype)
        self.DA_molecules = self.DA_test_res[self.DA_test_res["P-adjust"] <= self.threshold]["Entity"].tolist()
        self.results = []

    def over_representation_analysis(self):

        """
        Function for over representation analysis using Fisher exact test (right tailed)
        Returns:
            DataFrame of ORA results for each pathway, p-value, q-value, hits ratio
        """

        pathway_names = self.pathways["Pathway_name"].to_dict()
        pathway_dict = utils.pathwaydf_to_dict(self.pathways)

        # Remove pathways not present in the dataset
        compounds_present = self.DA_molecules
        pathways_present = {k: v for k, v in pathway_dict.items() if len([i for i in compounds_present if i in v]) >= 1}

        pathways_with_compounds = []
        pvalues = []
        pathway_ratio = []
        pathway_coverage = []

        for pathway in pathways_present:
            # perform ORA for each pathway
            pathway_compounds = pathway_dict[pathway]
            pathway_compounds = [i for i in pathway_compounds if str(i) != "nan"]
            if not pathway_compounds or len(pathway_compounds) < 2:
                # ignore pathway if contains no compounds or has less than 3 compounds
                continue
            else:
                DA_in_pathway = len(set(self.DA_molecules) & set(pathway_compounds))
                # k: compounds in DA list AND pathway
                DA_not_in_pathway = len(np.setdiff1d(self.DA_molecules, pathway_compounds))
                # K: compounds in DA list not in pathway
                compound_in_pathway_not_DA = len(set(pathway_compounds) & set(np.setdiff1d(self.background_set, self.DA_molecules)))
                # not DEM compounds present in pathway
                compound_not_in_pathway_not_DA = len(
                    np.setdiff1d(np.setdiff1d(self.background_set, self.DA_molecules), pathway_compounds))
                # compounds in background list not present in pathway
                if DA_in_pathway == 0 or (compound_in_pathway_not_DA + DA_in_pathway) < 2:
                    # ignore pathway if there are no DEM compounds in that pathway
                    continue
                else:
                    # Create 2 by 2 contingency table
                    pathway_ratio.append(str(DA_in_pathway) + "/" + str(compound_in_pathway_not_DA + DA_in_pathway))
                    pathway_coverage.append(
                        str(compound_in_pathway_not_DA + DA_in_pathway) + "/" + str(len(pathway_compounds)))
                    pathways_with_compounds.append(pathway)
                    contingency_table = np.array([[DA_in_pathway, compound_in_pathway_not_DA],
                                                [DA_not_in_pathway, compound_not_in_pathway_not_DA]])
                    # Run right tailed Fisher's exact test
                    oddsratio, pvalue = stats.fisher_exact(contingency_table, alternative="greater")
                    pvalues.append(pvalue)
        try:
            padj = sm.stats.multipletests(pvalues, 0.05, method="fdr_bh")
            results = pd.DataFrame(
                zip(pathways_with_compounds, pathway_ratio, pathway_coverage, pvalues,
                    padj[1]),
                columns=["ID",  "Hits", "Coverage", "P-value", "P-adjust"])
            results["Pathway_name"] = results["ID"].map(pathway_names)
            results.insert(1, 'Pathway_name', results.pop('Pathway_name'))

        except ZeroDivisionError:
            padj = [1] * len(pvalues)
            results = pd.DataFrame(zip(pathways_with_compounds, pathway_ratio, pvalues, padj),
                                columns=["ID", "Hits", "Coverage", "P-value", "P-adjust"])
            results["Pathway_name"] = results["ID"].map(pathway_names)
            results.insert(1, 'Pathway_name', results.pop('Pathway_name'))
            
        self.results = results
        return results


import requests
import pandas as pd
import numpy as np

def identifier_conversion(input_type, compound_list):
    """
    Use Metaboanalyst API for identifier conversion
    Args:
        input_type (str): identifier type present in input data - any of ('name', 'hmdb', 'pubchem', 'chebi', 'metlin', 'kegg')
        compound_list (list): list of identifiers in the data
    Returns:
        (pd.DataFrame) Dataframe containing identifier matches 
    """
    url = "http://api.xialab.ca/mapcompounds"

    resps = []
    for i in range(0, len(compound_list), 100):
        cpds = compound_list[i:i+100]
        input_cpds = {"queryList": cpds, "inputType": input_type}
        headers = {
            # 'Content-Type': "application/json",
            'cache-control': "no-cache",
            }

        response = requests.request("POST", url, json=input_cpds, headers=headers)
        resp_dict = response.json()
        resps.append(resp_dict)

    res_dict = {k: [d[k] for d in resps] for k in resps[0]}
    res_dict = {k: [item for sublist in v for item in sublist] for k, v in res_dict.items()}

    resp_df = pd.DataFrame(res_dict)
    return resp_df

def map_identifiers(query_df, output_id_type, matrix):
    """
    Map desired identifiers to input data
    Args:
        query_df (pd.DataFrame): DataFrame obtained using the identifier_conversion function containing ID mappings
        output_id_type (str): Any of ('Match', 'HMDB', 'PubChem', 'ChEBI', 'KEGG', 'METLIN','SMILES')
        matrix (pd.DataFrame): sample-by-compound metabolomics data matrix
    Returns:
        Sample-by-compound metabolomics data matrix with mapped identifiers, any compounds without a matching ID will be dropped
    """


    # output id type can be any of ['HMDB', 'PubChem', 'ChEBI', 'KEGG', 'METLIN','SMILES']
    cpd_mapping_dict = dict(zip(query_df["Query"].tolist(), query_df[output_id_type].tolist()))
    cpd_mapping_dict = {k: v for k, v in cpd_mapping_dict.items() if v not in ["NA", np.nan, None, "None"]}
    renamed_mat = matrix.drop([i for i in matrix.columns if i not in cpd_mapping_dict.keys()], axis=1)
    renamed_mat = renamed_mat.rename(cpd_mapping_dict, axis=1)
    return renamed_mat


from pkg_resources import get_distribution
__version__ = get_distribution('sspa').version

from .process_pathways import process_reactome, process_kegg, process_gmt
from .sspa_cluster import sspa_ssClustPA
from .sspa_kpca import sspa_kpca
from .sspa_zscore import sspa_zscore
from .sspa_svd import sspa_svd
from .utils import load_example_data, t_tests
from .sspa_ora import sspa_ora
from .sspa_gsea import sspa_gsea
from .sspa_ssGSEA import sspa_ssGSEA
from .download_pathways import download_KEGG, download_reactome, MetExplorePaths
from .identifier_conversion import identifier_conversion, map_identifiers


import numpy as np
import pandas as pd
import sspa.utils as utils
import gseapy

def sspa_gsea(mat, metadata, pathway_df, ranking_metric='signal_to_noise', min_entity=2):
    """Run GSEA using gseapy package by zqfang (https://github.com/zqfang/GSEApy)

    Args:
        mat (pd.DataFrame): dataframe containing input metabolomics data
        metadata (pd.Series): series containing phenotype metadata e.g 'COVID', 'NON-COVID'
        pathway_df (pd.DataFrame): GMT-like pathway dataframe containing compound identifiers
        ranking_metric (str): Ranking metric for molecules in GSEA. Default is signal-to-noise ratio. 
            Other options are 't_test' and see GSEApy package https://github.com/zqfang/GSEApy/blob/2b5419e14615b6fd19a575ff065256dc7099bbec/gseapy/gsea.py#L135 for more options. 
        min_entity (int, optional): minimum number of molecules mapping to pathways for GSEA to be performed. Defaults to 2.
    """
    
    pathway_names = pathway_df["Pathway_name"].to_dict()
    pathways = utils.pathwaydf_to_dict(pathway_df)
    compounds_present = mat.columns.tolist()
    pathways = {k: v for k, v in pathways.items() if len([i for i in compounds_present if i in v]) >= min_entity}

    gsea_res = gseapy.gsea(data=mat.T, 
                 gene_sets=pathways, 
                 cls=metadata,
                 min_size=min_entity,
                 permutation_type='phenotype',
                 permutation_num=1000, # reduce number to speed up test
                 outdir=None,  # do not write output to disk
                 method=ranking_metric)

    res_df = gsea_res.res2d
    res_df = res_df.rename(columns={'Term': 'Pathway_ID', 'NOM p-val': 'P-value', 'FDR q-val': 'P-adjust FDR', 'FWER p-val': 'P-adjust FWER',
     'Lead_genes': 'Leading_edge', 'Gene %': 'Entity %'})
    res_df = res_df.drop(['Name'], axis=1)
    name_col = res_df['Pathway_ID'].map(pathway_names)
    res_df.insert(1, 'Pathway_name', name_col)

    return res_df


import pandas as pd
import numpy as np
import sspa.utils as utils

def sspa_svd(mat, pathway_df, min_entity=2):

    """
    Tomfohr et al 2004 SVD/PLAGE method for single sample pathway analysis

    Args:
        mat (pd.DataFrame): pandas DataFrame omics data matrix consisting of m rows (samples) and n columns (entities).
        Do not include metadata columns
        pathways (pd.DataFrame): Dictionary of pathway identifiers (keys) and corresponding list of pathway entities (values).
        Entity identifiers must match those in the matrix columns
        min_entity (int): minimum number of metabolites mapping to pathways for ssPA to be performed

    Returns:
        pandas DataFrame of pathway scores derived using the PLAGE method. Columns represent pathways and rows represent samples.
    """

    pathways = utils.pathwaydf_to_dict(pathway_df)

    pathway_activities = []

    # Create pathway matrices
    pathway_ids = []
    for pathway, compounds in pathways.items():
        single_pathway_matrix = mat.drop(mat.columns.difference(compounds), axis=1)
        if single_pathway_matrix.shape[1] >= min_entity:
            pathway_ids.append(pathway)
            pathway_mat = single_pathway_matrix.T.values

            # s = singular values
            # u = left singular vector
            # v = right singular vector
            u, s, vh = np.linalg.svd(pathway_mat)

            pathway_activities.append(vh[0])

    pathway_activities_df = pd.DataFrame(pathway_activities, columns=mat.index, index=pathway_ids).T
    return pathway_activities_df


import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
import sspa.utils as utils


def sspa_ssClustPA(mat, pathway_df, min_entity=2):

    """
    K-means based clustering method for single sample pathway analysis

    Args:
        mat (pd.DataFrame): pandas DataFrame omics data matrix consisting of m rows (samples) and n columns (entities).
        Do not include metadata columns
        pathways (pd.DataFrame): Dictionary of pathway identifiers (keys) and corresponding list of pathway entities (values).
        Entity identifiers must match those in the matrix columns
        min_entity (int): minimum number of metabolites mapping to pathways for ssPA to be performed

    Returns:
        pandas DataFrame of pathway scores derived using the ssClustPA/(proj) method. Columns represent pathways and rows represent samples.
    """

    pathways = utils.pathwaydf_to_dict(pathway_df)

    pathway_matrices = []
    pathway_ids = []
    for pathway, compounds in pathways.items():
        single_pathway_matrix = mat.drop(mat.columns.difference(compounds), axis=1)
        if single_pathway_matrix.shape[1] >= min_entity:
            pathway_matrices.append(single_pathway_matrix.values)
            pathway_ids.append(pathway)

    scores = []
    for m in pathway_matrices:
        kmeans = KMeans(n_clusters=2).fit(m)
        centroids1 = kmeans.cluster_centers_[0]
        centroids2 = kmeans.cluster_centers_[1]

        vec = centroids1 - centroids2
        unit_vec = vec / np.linalg.norm(vec)
        proj_data = unit_vec.dot(m.T)
        scores.append(proj_data)

    scores_df = pd.DataFrame(scores, columns=mat.index, index=pathway_ids).T
    return scores_df


import numpy as np
import pandas as pd
import sspa.utils as utils
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri
from rpy2.robjects.conversion import localconverter

# for rpy2
base = importr('base')

def sspa_fgsea(mat, metadata, pathway_df, min_entity=2):
    """
    Function for gene/metabolite set enrichment analysis using fGSEA method (Korotkevich et al.)
    Args:
        mat (pd.DataFrame): dataframe containing input metabolomics data
        metadata (pd.Series): series containing phenotype metadata e.g 'COVID', 'NON-COVID'
        pathway_df (pd.DataFrame): pathway dataframe containing compound identifiers
        min_entity (int): minimum number of metabolites mapping to a pathway for it to be tested
    Returns:
        DataFrame of GSEA results for each pathway, p-value, q-value, ES, NES, leading egde genes/metabolites
    """

    pathway_names = pathway_df["Pathway_name"].to_dict()
    pathways = utils.pathwaydf_to_dict(pathway_df)
    compounds_present = mat.columns.tolist()
    pathways = {k: v for k, v in pathways.items() if len([i for i in compounds_present if i in v]) >= min_entity}

    # Get rankings - SNR
    mat['Target'] = pd.factorize(metadata)[0]
    # Check user has only input two classes of samples 
    if len(set(mat['Target'].tolist())) > 2:
        raise ValueError('More than two metadata classes detected. Only two metadata classes are supported in GSEA.')

    class_a = mat.loc[mat["Target"] == 0]
    class_a.drop(['Target'], axis=1, inplace=True)
    class_b = mat.loc[mat["Target"] != 0]
    class_b.drop(['Target'], axis=1, inplace=True)

    a_mean = np.mean(class_a, axis=0)
    b_mean = np.mean(class_b, axis=0)
    a_std = np.std(class_a, axis=0)
    b_std = np.std(class_b, axis=0)

    means = a_mean.subtract(b_mean)
    stds = a_std + b_std

    snr = means / stds
    snr_dict = snr.to_dict()
    
    with localconverter(ro.default_converter + pandas2ri.converter):
        r_list = ro.ListVector(pathways)  # pathways
        r_ranks = ro.ListVector(snr_dict) # ranks 

    ro.r('''
    unl <- function(v){
    test <- unlist(v)
    return(test)}
    ''')

    r_unl = ro.globalenv['unl']
    r_ranks = r_unl(r_ranks)

    fgsea_r = importr('fgsea')
    fgsea_res = fgsea_r.fgsea(pathways=r_list, stats=r_ranks)
    
    df = pd.DataFrame(fgsea_res)
    df = df.T
    df = df.apply(pd.to_numeric, errors="ignore")
    df.columns = ["ID", "P-value", "P-adjust", "log2err", "ES", "NES", "coverage", "leadingEdge"]
    df["Pathway_name"] = df["ID"].map(pathway_names)
    
    return df



# Cite Lee et al. 2008
import pandas as pd
import numpy as np
import scipy.stats as stats
import sspa.utils as utils

def sspa_zscore(mat, pathway_df, min_entity=2):
    """
    Lee at al 2008 z-score method for single sample pathway analysis

    Args:
        mat (pd.DataFrame): pandas DataFrame omics data matrix consisting of m rows (samples) and n columns (entities).
        Do not include metadata columns
        pathways (pd.DataFrame): Dictionary of pathway identifiers (keys) and corresponding list of pathway entities (values).
        Entity identifiers must match those in the matrix columns
        min_entity (int): minimum number of metabolites mapping to pathways for ssPA to be performed


    Returns:
        pandas DataFrame of pathway scores derived using the z-score method. Columns represent pathways and rows represent samples.
    """

    pathways = utils.pathwaydf_to_dict(pathway_df)

    pathway_activities = []
    pathway_ids = []
    for pathway, compounds in pathways.items():
        single_pathway_matrix = mat.drop(mat.columns.difference(compounds), axis=1)
        if single_pathway_matrix.shape[1] >= min_entity:
            pathway_mat = single_pathway_matrix.T.values
            pathway_ids.append(pathway)

            zscores = stats.zscore(pathway_mat, axis=1)
            sum_zscore = np.sum(zscores, axis=0)
            pathway_act = sum_zscore / np.sqrt(pathway_mat.shape[0])
            pathway_activities.append(pathway_act)

    pathway_activities_df = pd.DataFrame(pathway_activities, columns=mat.index, index=pathway_ids).T
    return pathway_activities_df




import pandas as pd
import sspa.utils as utils
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri
from rpy2.robjects.conversion import localconverter

# for rpy2
base = importr('base')

def sspa_gsva(mat, pathway_df, min_entity=2):

    """
    Hanzelmann et al GSVA method for single sample pathway analysis. 
    This is an rpy2 wrapper script to run the R implementation of GSVA.

    Args:
        mat (pd.DataFrame): pandas DataFrame omics data matrix consisting of m rows (samples) and n columns (entities).
        Do not include metadata columns
        pathways (pd.DataFrame): Dictionary of pathway identifiers (keys) and corresponding list of pathway entities (values).
        Entity identifiers must match those in the matrix columns
        min_entity (int): minimum number of metabolites mapping to pathways for ssPA to be performed

    Returns:
        pandas DataFrame of pathway scores derived using the GSVA method. Columns represent pathways and rows represent samples.
    """

    pathways = utils.pathwaydf_to_dict(pathway_df)
    compounds_present = mat.columns.tolist()
    pathways = {k: v for k, v in pathways.items() if len([i for i in compounds_present if i in v]) >= min_entity}

    with localconverter(ro.default_converter + pandas2ri.converter):
        r_mat = ro.conversion.py2rpy(mat.T)
    r_mat = base.as_matrix(r_mat)  # abundance matrix
    row_vec = base.as_character(mat.columns.tolist())
    r_mat.rownames = row_vec
    r_list = ro.ListVector(pathways)  # pathways
    gsva_r = importr('GSVA')
    gsva_res = gsva_r.gsva(r_mat, r_list)
    with localconverter(ro.default_converter + pandas2ri.converter):
        gsva_df = ro.conversion.rpy2py(gsva_res)
    gsva_res_df = pd.DataFrame(gsva_df.T, columns=pathways.keys(), index=mat.index.tolist())

    return gsva_res_df


import pandas as pd
import pkg_resources
import sspa.download_pathways 

def process_reactome(organism, infile=None, download_latest=False, filepath=None):
    '''
    Function to load Reactome pathways 
    Args:
        organism (str): Reactome organism name
        infile (str): default None, provide a Reactome pathway file to process into the GMT-style dataframe 
        download_latest (Bool): Downloads the latest version of Reactome metabolic pathways
        filepath (str): filepath to save pathway file to, default is None - save to variable
    Returns: 
        GMT-like pd.DataFrame containing Reactome pathways
    '''

    # Process CHEBI to reactome data

    if download_latest:
        pathways_df = sspa.download_pathways.download_reactome(organism, filepath)
        return pathways_df

    else:
        if infile == None or infile == "R78":
            stream = pkg_resources.resource_stream(__name__, 'pathway_databases/ChEBI2Reactome_All_Levels_R78.txt')
            f = pd.read_csv(stream, sep="\t", header=None, encoding='latin-1')
        else:
            f = pd.read_csv(infile, sep="\t", header=None)
        f.columns = ['CHEBI', 'pathway_ID', 'link', 'pathway_name', 'evidence_code', 'species']
        f_filt = f[f.species == organism]
        name_dict = dict(zip(f_filt['pathway_ID'], f_filt['pathway_name']))

        groups = f_filt.groupby(['pathway_ID'])['CHEBI'].apply(list).to_dict()
        groups = {k: list(set(v)) for k, v in groups.items()}
        df = pd.DataFrame.from_dict(groups, orient='index', dtype="object")
        pathways_df = df.dropna(axis=0, how='all', subset=df.columns.tolist()[1:])
        pathways_df = df.dropna(axis=1, how='all')

        pathways_df["Pathway_name"] = pathways_df.index.map(name_dict)
        pathways_df.insert(0, 'Pathway_name', pathways_df.pop('Pathway_name'))

        return pathways_df

def process_kegg(organism, infile=None, download_latest=False, filepath=None):
    '''
    Function to load KEGG pathways 
    Args:
        organism (str): KEGG organism code
        infile (str): default None, provide a KEGG pathway file to process into the GMT-style dataframe 
        download_latest (Bool): Downloads the latest version of KEGG metabolic pathways
        filepath (str): filepath to save pathway file to, default is None - save to variable
    Returns: 
        GMT-like pd.DataFrame containing KEGG pathways
    '''
    if download_latest:
        pathways_df = sspa.download_pathways.download_KEGG(organism, filepath)
        return pathways_df

    else:
        if infile == None or infile == "R98":
            stream = pkg_resources.resource_stream(__name__, 'pathway_databases/KEGG_human_pathways_compounds_R98.csv')
            pathways_df = pd.read_csv(stream, index_col=0, encoding='latin-1')
        else:
            pathways_df = pd.read_csv(infile, index_col=0)

        pathways_df = pathways_df.dropna(axis=0, how='all', subset=pathways_df.columns.tolist()[1:])
        pathways_df = pathways_df.dropna(axis=1, how='all')

        # Remove duplicated compounds
        mask = pathways_df.apply(pd.Series.duplicated, 1) & pathways_df.astype(bool)
        pathways_df = pathways_df.mask(mask, None)

        return pathways_df

def process_gmt(infile):
    '''
    Function to load pathways from a custom GMT-like file
    Args:
        infile (str): default None, provide a GMT pathway file to process into the GMT-style dataframe, file ending can be .csv or .gmt
    Returns: 
        GMT-like pd.DataFrame containing pathways
    '''
    if infile[-4:] == ".csv":
        pathways_df = pd.read_csv(infile, index_col=0)
    elif infile[-4:] == ".gmt":
        input_gmt = []
        with open(infile, "r") as f:
            for i in f:
                input_gmt.append(i.strip("\n").split("\t"))
        pathways_df = pd.DataFrame(input_gmt)
        pathways_df = pathways_df.rename({0:"Pathway_ID", 1:"Pathway_name"}, axis=1)
        pathways_df.index = pathways_df["Pathway_ID"]
        pathways_df = pathways_df.drop("Pathway_ID", axis=1)

    pathways_df = pathways_df.dropna(axis=0, how='all', subset=pathways_df.columns.tolist()[1:])
    pathways_df = pathways_df.dropna(axis=1, how='all')
    
    return pathways_df




import pandas as pd
import pkg_resources


class ProcessPathways:
    """
    Process pathway database files for use with sspa.

    :param infile : For Reactome, release 78 is available for all organisms and for KEGG release 98 (homo sapiens) is available.
     Otherwise, specify the full path to the pathway input file. For reactome pathways, download
     the ChEBI2Reactome_All_levels file from https://reactome.org/download-data and specify the full filepath to this.
    :param organism : the organism requried when using Reactome. Full name must be specified as in Reacome db e.g. ('Homo sapiens' or 'Danio rerio')

    :returns pathway entity dictionary (composed of pathway IDs and the entities each pathway is composed of),
     and pathway name mapping dict (composed of pathway ID: pathway name)
    """

    def __init__(self, infile, organism):
        self.infile = infile
        self.organism = organism

    def process_reactome(self):
        # Process CHEBI to reactome data

        if self.infile == "R78":
            stream = pkg_resources.resource_stream(__name__, 'pathway_databases/ChEBI2Reactome_All_Levels_R78.txt')
            f = pd.read_csv(stream, sep="\t", header=None, encoding='latin-1')
        else:
            f = pd.read_csv(self.infile, sep="\t", header=None)
        f.columns = ['CHEBI', 'pathway_ID', 'link', 'pathway_name', 'evidence_code', 'species']
        f_filt = f[f.species == self.organism]
        name_dict = dict(zip(f_filt['pathway_ID'], f_filt['pathway_name']))

        groups = f_filt.groupby(['pathway_ID'])['CHEBI'].apply(list).to_dict()
        df = pd.DataFrame.from_dict(groups, orient='index', dtype="object")

        pathways_df = df.dropna(axis=0, how='all', subset=df.columns.tolist()[1:])
        pathways = pathways_df.index.tolist()
        pathway_dict = {}

        for pathway in pathways:
            pathway_compounds = list(set(pathways_df.loc[pathway, :].tolist()))
            pathway_compounds = [str(i) for i in pathway_compounds if str(i) != "None"]

            cpds = pathway_compounds[1:]
            if len(cpds) > 1:
                pathway_dict[pathway] = cpds
        return pathway_dict, name_dict

    def process_kegg(self):
        if self.infile == "R98":
            stream = pkg_resources.resource_stream(__name__, 'pathway_databases/KEGG_human_pathways_compounds_R98.csv')
            f = pd.read_csv(stream, index_col=0, encoding='latin-1')
        else:
            f = pd.read_csv(self.infile, index_col=0)
        name_dict = dict(zip(f.index, f['Pathway_name']))
        pathway_dict = {k: list(set(f.loc[k, '0':].tolist())) for k in list(name_dict.keys())}
        pathway_dict = {k: [i for i in v if pd.notnull(i)] for k, v in pathway_dict.items()}
        pathway_dict = {k: v for k, v in pathway_dict.items() if len(v) > 2}

        return pathway_dict, name_dict
        # remove dupes



import pandas as pd
import numpy as np
import pkg_resources
import scipy.stats as stats
import statsmodels.api as sm

def load_example_data(omicstype="metabolomics", processed=True):
    """
    Loads example datasets

    Args:
        omicstype (str): type of omics for example data. 
            Available options are "metabolomics" or "transcriptomics". 
            Metabolomics data are from Su et al 2020 https://doi.org/10.1016/j.cell.2020.10.037.
            Transcriptomics data - TO BE IMPLEMENTED
        processed (bool): Load processed (normalised, scaled) or raw data

    Returns:
        pre-processed omics data matrix consisting of m samples and n entities (metabolites/genes) in the form of a pandas DataFrame. 
        Contains one of more metadata columns at the end.
    """

    if omicstype == "metabolomics":
        if processed:
            stream = pkg_resources.resource_stream(__name__, 'example_data/Su_covid_metabolomics_processed.csv')
            f = pd.read_csv(stream, index_col=0, encoding='latin-1')
            return f
        else:
            stream = pkg_resources.resource_stream(__name__, 'example_data/Su_metab_data_raw.csv')
            f = pd.read_csv(stream, index_col=0, encoding='latin-1')
            return f


def t_tests(matrix, classes, multiple_correction_method, testtype="ttest"):
    """
    Performs two-sample independent t-tests

    Args:
        matrix (pd.DataFrame): processed sample-by-compound metabolomics dataframe
        classes (pd.Series): pandas series containing phenotype metadata (e.g. 'COVID', 'NON-COVID')
        multiple_correction_method (str): see https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html for options
        testtype (str): Default is t-test, "mwu" also available to implement the Mann Whitney U test

    Returns:
        pd.DataFrame containing p-values and corrected p-values for each metabolite
    """
    metabolites = matrix.columns.tolist()
    matrix['Target'] = pd.factorize(classes)[0]

    # Check user has only input two classes of samples 
    if len(set(matrix['Target'].tolist())) > 2:
        raise ValueError('More than two metadata classes detected. Only two metadata classes are supported in ORA.')

    disease = matrix[matrix["Target"] == 0]
    disease = disease.drop(['Target'], axis=1)
    ctrl = matrix[matrix["Target"] != 0]
    ctrl = ctrl.drop(['Target'], axis=1)
    matrix = matrix.drop(['Target'], axis=1)

    # disease = matrix.loc[matrix["Target"] == 0]
    # disease.drop(['Target'], axis=1, inplace=True)
    # ctrl = matrix.loc[matrix["Target"] != 0]
    # ctrl.drop(['Target'], axis=1, inplace=True)
    # matrix = matrix.drop(['Target'], axis=1)
    
    if testtype == "mwu":
        pvalues = stats.mannwhitneyu(disease, ctrl, axis=0)[1]
    else:
        pvalues = stats.ttest_ind(disease, ctrl)[1]

    padj = sm.stats.multipletests(pvalues, 0.05, method=multiple_correction_method)
    results = pd.DataFrame(zip(metabolites, pvalues, padj[1]),
                           columns=["Entity", "P-value", "P-adjust"])
    return results


def pathwaydf_to_dict(df):
    """
    Converts pathway dataframe to dictionary, with pathway IDs as keys and metabolite lists as values
    Args:
        df (pd.DataFrame): Pandas DataFrame containing pathways 
    Returns: 
        python dict pathway representation
    """
    pathways_df = df.drop(["Pathway_name"], axis=1)
    pathway_dict = {}

    for pathway in pathways_df.index:
        pathway_compounds = list(set(pathways_df.loc[pathway, :].tolist()))
        pathway_compounds = [str(i) for i in pathway_compounds if str(i) not in ["None", np.nan, 'nan']]

        if len(pathway_compounds) > 1:
            pathway_dict[pathway] = pathway_compounds
    return pathway_dict


# Get latest KEGG metabolic pathways using rest API
# Similar to KEGGREST R in principle

import requests
import re
import pandas as pd
import warnings
import json

def download_KEGG(organism, filepath=None):
    '''
    Function for KEGG pathway download
    Args:
        organism (str): KEGG 3 letter organism code
        filepath (str): filepath to save pathway file to, default is None - save to variable
    Returns: 
        GMT-like pd.DataFrame containing KEGG pathways
    '''
    print("Beginning KEGG download...")
    # get all pathways
    url = 'http://rest.kegg.jp/list/pathway/'+organism
    # change organism name
    data = requests.get(url)
    pathways = data.text
    pathways = pathways.split("\n")
    pathways = filter(None, pathways)
    pathway_dict = dict()

    for path in pathways:
        path = path.split("\t")
        name = path[1]
        pathid = re.search(r"path:(.*)", path[0]).group(1)
        pathway_dict[pathid] = name

    # get compounds for each pathway
    base_url = 'http://rest.kegg.jp/get/'

    pathway_ids = [*pathway_dict]
    pathway_names = list(pathway_dict.values())
    pathway_compound_mapping = dict()

    for i in pathway_ids:
        complist = []
        current_url = base_url + "pathway:" + i

        # parse the pathway description page
        page = requests.get(current_url)
        lines = page.text.split("\n")

        try:
            cpds_start = [lines.index(i) for i in lines if i.startswith("COMPOUND")][0]
            reference_start = [lines.index(i) for i in lines if i.startswith("REFERENCE") or i.startswith("REL_PATHWAY")][0]
            cpds_lines = lines[cpds_start:reference_start]
            first_cpd = cpds_lines.pop(0).split()[1]
            complist.append(first_cpd)
            complist = complist + [i.split()[0] for i in cpds_lines]
            pathway_compound_mapping[i] = list(set(complist))
        except IndexError:
            pathway_compound_mapping[i] = []

    # get release details
    release_data = requests.get('http://rest.kegg.jp/info/kegg')
    version_no = release_data.text.split()[9][0:3]

    # create GMT style file
    df = pd.DataFrame.from_dict(pathway_compound_mapping, orient='index')
    df.insert(0, 'Pathway_name', pathway_names)

    if filepath:
        fpath = filepath + "/KEGG_" + organism + "_pathways_compounds_R" + str(version_no) + ".gmt"
        df.to_csv(fpath, sep="\t", header=False)
        print("KEGG DB file saved to " + fpath)
    print("Complete!")

    return df

def download_reactome(organism, filepath=None):
    '''
    Function for Reactome pathway download
    Args:
        organism (str): Reactome organism name
        filepath (str): filepath (str): filepath to save pathway file to, default is None - save to variable
    Returns: 
        GMT-like pd.DataFrame containing Reactome pathways
    '''
    print("Beginning Reactome download...")

     # get all pathways
    url = 'https://reactome.org/download/current/ChEBI2Reactome_All_Levels.txt'
    f = pd.read_csv(url, sep="\t", header=None)
    
    f.columns = ['CHEBI', 'pathway_ID', 'link', 'pathway_name', 'evidence_code', 'species']
    f_filt = f[f.species == organism]
    name_dict = dict(zip(f_filt['pathway_ID'], f_filt['pathway_name']))

    groups = f_filt.groupby(['pathway_ID'])['CHEBI'].apply(list).to_dict()
    groups = {k: list(set(v)) for k, v in groups.items()}

    df = pd.DataFrame.from_dict(groups, orient='index', dtype="object")
    pathways_df = df.dropna(axis=0, how='all', subset=df.columns.tolist()[1:])
    pathways_df = df.dropna(axis=1, how='all')

    pathways_df["Pathway_name"] = pathways_df.index.map(name_dict)
    pathways_df.insert(0, 'Pathway_name', pathways_df.pop('Pathway_name'))

    # get release details
    release_data = requests.get('https://reactome.org/download/current/reactome_stable_ids.txt')
    version_no = release_data.text.split()[6]

    if filepath:
        fpath = filepath + "/Reactome_" + "_".join(organism.split())+ "_pathways_compounds_R" + str(version_no) + ".gmt"
        pathways_df.to_csv(fpath, sep="\t", header=False)
        print("Reactome DB file saved to " + fpath)
    print("Complete!")

    return pathways_df

class MetExplorePaths:
    '''
    Class for downloading metexplore metabolic models in the form of pathways with mapped identifiers

    Attributes:
        model (str): identifier of genome scale metabolic model available on MetExplore
        id_type (str): identifier type for the model pathways
        filepath (str): filepath to save the pathway file to, default is None (save to variable)
        nMappedID (int): Number of metabolites mapping to the selected identifier type
        nMetab (int): Number of metabolites in the model
        pathways (pd.DataFrame): GMT-like format pathway DataFrame

    '''
    def __init__(self, model, id_type, filepath=None):
        self.model = model
        self.id_type = id_type
        self.filepath = filepath
        self.nMappedID = None
        self.nMetab = None
        self.pathways = None
        # downloads pathways on object instantiation
        self.download_metexplore()

    def download_metexplore(self):
        '''
        Function to download MetExplore pathways
        '''
        warnings.filterwarnings("ignore")
        metexploreURL = "https://metexplore.toulouse.inrae.fr/metexplore-api/"+str(self.model)+"/pathwaymetabolite/"+str(self.id_type)+"/"
        stats_nmapped_url = "https://metexplore.toulouse.inrae.fr/metexplore-api/stat/"+str(self.model)+"/"+str(self.id_type)+"/"
        stats_nmetab_url = "https://metexplore.toulouse.inrae.fr/metexplore-api/stat/"+str(self.model)+"/nbMetab/"
        
        stats_nmetab = requests.get(stats_nmetab_url, verify=False)
        stats_nmapped = requests.get(stats_nmapped_url, verify=False)
        data_api = requests.get(metexploreURL, verify=False)
        pathways_json = data_api.json()

        pathways_df = pd.DataFrame.from_dict(pathways_json, orient='columns')
        del(pathways_df["id"]) 
        
        pathways = pathways_df.merge(pathways_df.Metabolites.apply(pd.Series), right_index=True, left_index=True)
        del(pathways["Metabolites"])
        
        pathways = pathways.fillna("None")
        pathways.rename(columns={'name':"Pathway_name"}, inplace = True)
        
        pathways = pathways.fillna("None")
        pathways = pathways.set_index('dbIdentifier')
        pathways = pathways.drop(columns=['len'])
        pathways.rename(columns={'name':"Pathway_name"}, inplace = True)

        #if file path provided save gmt to drive
        if self.filepath:
            fpath = self.filepath + "/MetExplorePathways_" + str(self.model) + "_" + str(self.id_type) + ".gmt"
            pathways.to_csv(fpath, sep="\t", header=False)
            print("MetExplore metabolic network pathways file saved to " + fpath)

        self.pathways = pathways
        self.nMappedID = stats_nmapped.text.split("\n")[2]
        self.nMetab = stats_nmetab.text.split("\n")[2]
        
        print("Complete!")
        return pathways



from setuptools import setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='sspa',
    version='0.2.1',
    packages=['sspa'],
    package_dir={'':'src'},
    url='https://github.com/cwieder/sspa',
    license='GNU 3.0',
    author='Cecilia Wieder',
    author_email='cw2019@ic.ac.uk',
    description='Single sample pathway analysis tools for omics data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data={'sspa': ['example_data/*', 'pathway_databases/*']},
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ]
)



import pandas as pd
import sspa.utils as utils
import gseapy
# import rpy2.robjects as ro
# from rpy2.robjects.packages import importr
# from rpy2.robjects import pandas2ri
# from rpy2.robjects.conversion import localconverter

# # for rpy2
# base = importr('base')

def sspa_ssGSEA(mat, pathway_df, min_entity=2):

    """
    Barbie et al ssGSEA method for single sample pathway analysis. 
    Uses the ssgsea function of the gseapy package (https://github.com/zqfang/GSEApy) as a backend. 

    Args:
        mat (pd.DataFrame): pandas DataFrame omics data matrix consisting of m rows (samples) and n columns (entities).
        Do not include metadata columns
        pathways (pd.DataFrame): Dictionary of pathway identifiers (keys) and corresponding list of pathway entities (values).
        Entity identifiers must match those in the matrix columns
        min_entity (int): minimum number of metabolites mapping to pathways for ssPA to be performed


    Returns:
        pandas DataFrame of pathway scores derived using the ssGSEA method. Columns represent pathways and rows represent samples.
    """

    pathways = utils.pathwaydf_to_dict(pathway_df)
    compounds_present = mat.columns.tolist()
    pathways = {k: v for k, v in pathways.items() if len([i for i in compounds_present if i in v]) >= min_entity}

    ssgsea_res = gseapy.ssgsea(data=mat.T,
               gene_sets=pathways,
               min_size=min_entity,
               outdir=None,
               sample_norm_method='rank', # choose 'custom' will only use the raw value of `data`
               no_plot=True)

    ssgsea_scores = ssgsea_res.res2d.pivot(index='Term', columns='Name', values='NES').T
    res_df = pd.DataFrame(ssgsea_scores, index=mat.index, columns=pathways.keys())
    res_df = res_df.astype(float)
    return res_df

    # with localconverter(ro.default_converter + pandas2ri.converter):
    #     r_mat = ro.conversion.py2rpy(mat.T)
    # r_mat = base.as_matrix(r_mat)  # abundance matrix
    # row_vec = base.as_character(mat.columns.tolist())
    # r_mat.rownames = row_vec
    # r_list = ro.ListVector(pathways)  # pathways
    # gsva_r = importr('GSVA')
    # ssgsea_res = gsva_r.gsva(r_mat, r_list, method='ssgsea')
    # with localconverter(ro.default_converter + pandas2ri.converter):
    #     ssgsea_df = ro.conversion.rpy2py(ssgsea_res)
    # ssgsea_res_df = pd.DataFrame(ssgsea_df.T, columns=pathways.keys(), index=mat.index.tolist())

    # return ssgsea_res_df